@@ -1,5 +1,7 @@
+#include "base/bits.h"
 #include "caching/superblock.h"
 
+using namespace base;
 using namespace caching;
 using namespace superblock_damage;
 
@@ -43,6 +45,8 @@ namespace {
 		le32 write_misses;
 
 		le32 policy_version[CACHE_POLICY_VERSION_SIZE];
+
+		le64 dirty_root; // format 2 only
 	} __attribute__ ((packed));
 
 	struct superblock_traits {
@@ -55,7 +59,7 @@ namespace {
 
 	uint32_t const SUPERBLOCK_MAGIC = 06142003;
 	uint32_t const VERSION_BEGIN = 1;
-	uint32_t const VERSION_END = 2;
+	uint32_t const VERSION_END = 3;
 }
 
 //----------------------------------------------------------------
@@ -72,6 +76,11 @@ superblock_flags::superblock_flags(uint32_t bits)
 		bits &= ~(1 << CLEAN_SHUTDOWN_BIT);
 	}
 
+	if (bits & (1u << NEEDS_CHECK_BIT)) {
+		flags_.insert(NEEDS_CHECK);
+		bits &= ~(1u << NEEDS_CHECK_BIT);
+	}
+
 	unhandled_flags_ = bits;
 }
 
@@ -81,6 +90,12 @@ superblock_flags::set_flag(superblock_flags::flag f)
 	flags_.insert(f);
 }
 
+void
+superblock_flags::clear_flag(superblock_flags::flag f)
+{
+	flags_.erase(f);
+}
+
 bool
 superblock_flags::get_flag(flag f) const
 {
@@ -95,6 +110,9 @@ superblock_flags::encode() const
 	if (get_flag(CLEAN_SHUTDOWN))
 		r = r | (1 << CLEAN_SHUTDOWN_BIT);
 
+	if (get_flag(NEEDS_CHECK))
+		r = r | (1u << NEEDS_CHECK_BIT);
+
 	return r;
 }
 
@@ -111,7 +129,7 @@ superblock::superblock()
 	  blocknr(SUPERBLOCK_LOCATION),
 	  magic(SUPERBLOCK_MAGIC),
 	  version(VERSION_END - 1u),
-	  policy_hint_size(0),
+	  policy_hint_size(4),
 	  mapping_root(0),
 	  hint_root(0),
 	  discard_root(0),
@@ -139,6 +157,10 @@ superblock::superblock()
 void
 superblock_traits::unpack(superblock_disk const &disk, superblock &core)
 {
+	core.compat_flags = to_cpu<uint32_t>(disk.compat_flags);
+	core.compat_ro_flags = to_cpu<uint32_t>(disk.compat_ro_flags);
+	core.incompat_flags = to_cpu<uint32_t>(disk.incompat_flags);
+
 	core.csum = to_cpu<uint32_t>(disk.csum);
 
 	core.flags = superblock_flags(to_cpu<uint32_t>(disk.flags));
@@ -153,7 +175,8 @@ superblock_traits::unpack(superblock_disk const &disk, superblock &core)
 	for (unsigned i = 0; i < CACHE_POLICY_VERSION_SIZE; i++)
 		core.policy_version[i] = to_cpu<uint32_t>(disk.policy_version[i]);
 
-	core.policy_hint_size = core.version == 1 ? 4 : to_cpu<uint32_t>(disk.policy_hint_size);
+	core.policy_hint_size = test_bit(core.incompat_flags, VARIABLE_HINT_SIZE_BIT) ?
+		to_cpu<uint32_t>(disk.policy_hint_size) : 4;
 
 	::memcpy(core.metadata_space_map_root,
 		 disk.metadata_space_map_root,
@@ -170,19 +193,21 @@ superblock_traits::unpack(superblock_disk const &disk, superblock &core)
 	core.metadata_block_size = to_cpu<uint32_t>(disk.metadata_block_size);
 	core.cache_blocks = to_cpu<uint32_t>(disk.cache_blocks);
 
-	core.compat_flags = to_cpu<uint32_t>(disk.compat_flags);
-	core.compat_ro_flags = to_cpu<uint32_t>(disk.compat_ro_flags);
-	core.incompat_flags = to_cpu<uint32_t>(disk.incompat_flags);
-
 	core.read_hits = to_cpu<uint32_t>(disk.read_hits);
 	core.read_misses = to_cpu<uint32_t>(disk.read_misses);
 	core.write_hits = to_cpu<uint32_t>(disk.write_hits);
 	core.write_misses = to_cpu<uint32_t>(disk.write_misses);
+
+	if (core.version >= 2)
+		core.dirty_root = to_cpu<uint64_t>(disk.dirty_root);
 }
 
 void
-superblock_traits::pack(superblock const &core, superblock_disk &disk)
+superblock_traits::pack(superblock const &sb, superblock_disk &disk)
 {
+	// We adjust some of the flags in the superblock, so make a copy
+	superblock core(sb);
+
 	disk.csum = to_disk<le32>(core.csum);
 	disk.flags = to_disk<le32>(core.flags.encode());
 	disk.blocknr = to_disk<le64>(core.blocknr);
@@ -195,7 +220,14 @@ superblock_traits::pack(superblock const &core, superblock_disk &disk)
 
 	for (unsigned i = 0; i < CACHE_POLICY_VERSION_SIZE; i++)
 		disk.policy_version[i] = to_disk<le32>(core.policy_version[i]);
-	disk.policy_hint_size = to_disk<le32>(core.policy_hint_size);
+
+	if (core.policy_hint_size != 4) {
+		set_bit(core.incompat_flags, VARIABLE_HINT_SIZE_BIT);
+		disk.policy_hint_size = to_disk<le32>(core.policy_hint_size);
+	} else {
+		clear_bit(core.incompat_flags, VARIABLE_HINT_SIZE_BIT);
+		disk.policy_hint_size = to_disk<le32>(0u);
+	}
 
 	::memcpy(disk.metadata_space_map_root,
 		 core.metadata_space_map_root,
@@ -220,6 +252,11 @@ superblock_traits::pack(superblock const &core, superblock_disk &disk)
 	disk.read_misses = to_disk<le32>(core.read_misses);
 	disk.write_hits = to_disk<le32>(core.write_hits);
 	disk.write_misses = to_disk<le32>(core.write_misses);
+
+	// The version may be overridden, meaning the dirty root may not
+	// actually be present.
+	if (core.version >= 2 && core.dirty_root)
+		disk.dirty_root = to_disk<le64>(*core.dirty_root);
 }
 
 //--------------------------------
@@ -256,25 +293,34 @@ namespace validator {
         unsigned const SECTOR_TO_BLOCK_SHIFT = 3;
 	uint32_t const SUPERBLOCK_CSUM_SEED = 9031977;
 
-	struct sb_validator : public block_manager<>::validator {
-		virtual void check(buffer<> const &b, block_address location) const {
-			superblock_disk const *sbd = reinterpret_cast<superblock_disk const *>(&b);
+	struct sb_validator : public bcache::validator {
+		virtual void check(void const *raw, block_address location) const {
+			superblock_disk const *sbd = reinterpret_cast<superblock_disk const *>(raw);
 			crc32c sum(SUPERBLOCK_CSUM_SEED);
 			sum.append(&sbd->flags, MD_BLOCK_SIZE - sizeof(uint32_t));
 			if (sum.get_sum() != to_cpu<uint32_t>(sbd->csum))
 				throw checksum_error("bad checksum in superblock");
 		}
 
-		virtual void prepare(buffer<> &b, block_address location) const {
-			superblock_disk *sbd = reinterpret_cast<superblock_disk *>(&b);
+		virtual bool check_raw(void const *raw) const {
+			superblock_disk const *sbd = reinterpret_cast<superblock_disk const *>(raw);
+			crc32c sum(SUPERBLOCK_CSUM_SEED);
+			sum.append(&sbd->flags, MD_BLOCK_SIZE - sizeof(uint32_t));
+			if (sum.get_sum() != to_cpu<uint32_t>(sbd->csum))
+				return false;
+			return true;
+		}
+
+		virtual void prepare(void *raw, block_address location) const {
+			superblock_disk *sbd = reinterpret_cast<superblock_disk *>(raw);
 			crc32c sum(SUPERBLOCK_CSUM_SEED);
 			sum.append(&sbd->flags, MD_BLOCK_SIZE - sizeof(uint32_t));
 			sbd->csum = to_disk<base::le32>(sum.get_sum());
 		}
 	};
 
-	block_manager<>::validator::ptr  mk_v() {
-		return block_manager<>::validator::ptr(new sb_validator);
+	bcache::validator::ptr  mk_v() {
+		return bcache::validator::ptr(new sb_validator);
 	}
 }
 
@@ -283,9 +329,10 @@ namespace validator {
 superblock
 caching::read_superblock(block_manager<>::ptr bm, block_address location)
 {
+	using namespace validator;
 	superblock sb;
-	block_manager<>::read_ref r = bm->read_lock(location, validator::mk_v());
-	superblock_disk const *sbd = reinterpret_cast<superblock_disk const *>(&r.data());
+	block_manager<>::read_ref r = bm->read_lock(location, mk_v());
+	superblock_disk const *sbd = reinterpret_cast<superblock_disk const *>(r.data());
 	superblock_traits::unpack(*sbd, sb);
 
 	return sb;
@@ -294,8 +341,9 @@ caching::read_superblock(block_manager<>::ptr bm, block_address location)
 void
 caching::write_superblock(block_manager<>::ptr bm, superblock const &sb, block_address location)
 {
-	block_manager<>::write_ref w = bm->superblock_zero(location, validator::mk_v());
-	superblock_traits::pack(sb, *reinterpret_cast<superblock_disk *>(w.data().raw()));
+	using namespace validator;
+	block_manager<>::write_ref w = bm->superblock_zero(location, mk_v());
+	superblock_traits::pack(sb, *reinterpret_cast<superblock_disk *>(w.data()));
 }
 
 void
@@ -361,7 +409,7 @@ caching::check_superblock(superblock const &sb,
 		visitor.visit(superblock_invalid(msg.str()));
 	}
 
-	if (sb.incompat_flags != 0) {
+	if (sb.incompat_flags & ~(1 << VARIABLE_HINT_SIZE_BIT)) {
 		ostringstream msg;
 		msg << "incompat_flags invalid (can only be 0): " << sb.incompat_flags;
 		visitor.visit(superblock_invalid(msg.str()));
