@@ -16,13 +16,16 @@
 // with thin-provisioning-tools.  If not, see
 // <http://www.gnu.org/licenses/>.
 
+#include "base/endian_utils.h"
+
 #include "persistent-data/space-maps/disk.h"
 #include "persistent-data/space-maps/disk_structures.h"
 #include "persistent-data/space-maps/recursive.h"
 #include "persistent-data/space-maps/careful_alloc.h"
 
+#include "persistent-data/data-structures/btree_damage_visitor.h"
+#include "persistent-data/data-structures/btree_counter.h"
 #include "persistent-data/checksum.h"
-#include "persistent-data/endian_utils.h"
 #include "persistent-data/math_utils.h"
 #include "persistent-data/transaction_manager.h"
 
@@ -35,9 +38,9 @@ using namespace sm_disk_detail;
 namespace {
 	uint64_t const BITMAP_CSUM_XOR = 240779;
 
-	struct bitmap_block_validator : public block_manager<>::validator {
-		virtual void check(buffer<> const &b, block_address location) const {
-			bitmap_header const *data = reinterpret_cast<bitmap_header const *>(&b);
+	struct bitmap_block_validator : public bcache::validator {
+		virtual void check(void const *raw, block_address location) const {
+			bitmap_header const *data = reinterpret_cast<bitmap_header const *>(raw);
 			crc32c sum(BITMAP_CSUM_XOR);
 			sum.append(&data->not_used, MD_BLOCK_SIZE - sizeof(uint32_t));
 			if (sum.get_sum() != to_cpu<uint32_t>(data->csum))
@@ -47,8 +50,17 @@ namespace {
 				throw checksum_error("bad block nr in space map bitmap");
 		}
 
-		virtual void prepare(buffer<> &b, block_address location) const {
-			bitmap_header *data = reinterpret_cast<bitmap_header *>(&b);
+		virtual bool check_raw(void const *raw) const {
+			bitmap_header const *data = reinterpret_cast<bitmap_header const *>(raw);
+			crc32c sum(BITMAP_CSUM_XOR);
+			sum.append(&data->not_used, MD_BLOCK_SIZE - sizeof(uint32_t));
+			if (sum.get_sum() != to_cpu<uint32_t>(data->csum))
+				return false;
+			return true;
+		}
+
+		virtual void prepare(void *raw, block_address location) const {
+			bitmap_header *data = reinterpret_cast<bitmap_header *>(raw);
 			data->blocknr = to_disk<base::le64, uint64_t>(location);
 
 			crc32c sum(BITMAP_CSUM_XOR);
@@ -59,12 +71,10 @@ namespace {
 
 	//--------------------------------
 
-	uint64_t const INDEX_CSUM_XOR = 160478;
-
 	// FIXME: factor out the common code in these validators
-	struct index_block_validator : public block_manager<>::validator {
-		virtual void check(buffer<> const &b, block_address location) const {
-			metadata_index const *mi = reinterpret_cast<metadata_index const *>(&b);
+	struct index_block_validator : public bcache::validator {
+		virtual void check(void const *raw, block_address location) const {
+			metadata_index const *mi = reinterpret_cast<metadata_index const *>(raw);
 			crc32c sum(INDEX_CSUM_XOR);
 			sum.append(&mi->padding_, MD_BLOCK_SIZE - sizeof(uint32_t));
 			if (sum.get_sum() != to_cpu<uint32_t>(mi->csum_))
@@ -74,8 +84,17 @@ namespace {
 				throw checksum_error("bad block nr in metadata index block");
 		}
 
-		virtual void prepare(buffer<> &b, block_address location) const {
-			metadata_index *mi = reinterpret_cast<metadata_index *>(&b);
+		virtual bool check_raw(void const *raw) const {
+			metadata_index const *mi = reinterpret_cast<metadata_index const *>(raw);
+			crc32c sum(INDEX_CSUM_XOR);
+			sum.append(&mi->padding_, MD_BLOCK_SIZE - sizeof(uint32_t));
+			if (sum.get_sum() != to_cpu<uint32_t>(mi->csum_))
+				return false;
+			return true;
+		}
+
+		virtual void prepare(void *raw, block_address location) const {
+			metadata_index *mi = reinterpret_cast<metadata_index *>(raw);
 			mi->blocknr_ = to_disk<base::le64, uint64_t>(location);
 
 			crc32c sum(INDEX_CSUM_XOR);
@@ -84,11 +103,6 @@ namespace {
 		}
 	};
 
-	block_manager<>::validator::ptr
-	index_validator() {
-		return block_manager<>::validator::ptr(new index_block_validator());
-	}
-
 	//--------------------------------
 
 	class bitmap {
@@ -96,26 +110,26 @@ namespace {
 		typedef transaction_manager::read_ref read_ref;
 		typedef transaction_manager::write_ref write_ref;
 
-		bitmap(transaction_manager::ptr tm,
+		bitmap(transaction_manager &tm,
 		       index_entry const &ie,
-		       block_manager<>::validator::ptr v)
+		       bcache::validator::ptr v)
 			: tm_(tm),
 			  validator_(v),
 			  ie_(ie) {
 		}
 
 		ref_t lookup(unsigned b) const {
-			read_ref rr = tm_->read_lock(ie_.blocknr_, validator_);
+			read_ref rr = tm_.read_lock(ie_.blocknr_, validator_);
 			void const *bits = bitmap_data(rr);
 			ref_t b1 = test_bit_le(bits, b * 2);
 			ref_t b2 = test_bit_le(bits, b * 2 + 1);
 			ref_t result = b2 ? 1 : 0;
-			result |= b1 ? 0b10 : 0;
+			result |= b1 ? 2 : 0;
 			return result;
 		}
 
 		void insert(unsigned b, ref_t n) {
-			write_ref wr = tm_->shadow(ie_.blocknr_, validator_).first;
+			write_ref wr = tm_.shadow(ie_.blocknr_, validator_).first;
 			void *bits = bitmap_data(wr);
 			bool was_free = !test_bit_le(bits, b * 2) && !test_bit_le(bits, b * 2 + 1);
 			if (n == 1 || n == 3)
@@ -156,31 +170,31 @@ namespace {
 		}
 
 		void iterate(block_address offset, block_address hi, space_map::iterator &it) const {
-			read_ref rr = tm_->read_lock(ie_.blocknr_, validator_);
+			read_ref rr = tm_.read_lock(ie_.blocknr_, validator_);
 			void const *bits = bitmap_data(rr);
 
 			for (unsigned b = 0; b < hi; b++) {
 				ref_t b1 = test_bit_le(bits, b * 2);
 				ref_t b2 = test_bit_le(bits, b * 2 + 1);
 				ref_t result = b2 ? 1 : 0;
-				result |= b1 ? 0b10 : 0;
+				result |= b1 ? 2 : 0;
 				it(offset + b, result);
 			}
 		}
 
 	private:
 		void *bitmap_data(transaction_manager::write_ref &wr) {
-			bitmap_header *h = reinterpret_cast<bitmap_header *>(&wr.data()[0]);
+			bitmap_header *h = reinterpret_cast<bitmap_header *>(wr.data());
 			return h + 1;
 		}
 
 		void const *bitmap_data(transaction_manager::read_ref &rr) const {
-			bitmap_header const *h = reinterpret_cast<bitmap_header const *>(&rr.data()[0]);
+			bitmap_header const *h = reinterpret_cast<bitmap_header const *>(rr.data());
 			return h + 1;
 		}
 
-		transaction_manager::ptr tm_;
-		block_manager<>::validator::ptr validator_;
+		transaction_manager &tm_;
+		bcache::validator::ptr validator_;
 
 		index_entry ie_;
 	};
@@ -199,6 +213,7 @@ namespace {
 		}
 	};
 
+#if 0
 	class ref_count_checker : public btree_checker<1, ref_count_traits> {
 	public:
 		typedef boost::shared_ptr<ref_count_checker> ptr;
@@ -207,18 +222,27 @@ namespace {
 			: btree_checker<1, ref_count_traits>(counter) {
 		}
 	};
+#endif
+
+	class index_entry_visitor {
+	public:
+		virtual ~index_entry_visitor() {}
+		virtual void visit(index_entry const &ie) = 0;
+		virtual void visit(run<block_address> const &missing) = 0;
+	};
 
 	class index_store {
 	public:
 		typedef boost::shared_ptr<index_store> ptr;
 
+		virtual void count_metadata(block_counter &bc) const = 0;
 		virtual void resize(block_address nr_indexes) = 0;
 		virtual index_entry find_ie(block_address b) const = 0;
 		virtual void save_ie(block_address b, struct index_entry ie) = 0;
 		virtual void commit_ies() = 0;
 		virtual ptr clone() const = 0;
 		virtual block_address get_root() const = 0;
-		virtual void check(block_counter &counter, block_address nr_index_entries) const = 0;
+		virtual void visit(index_entry_visitor &v, block_address nr_index_entries) const = 0;
 	};
 
 	unsigned const ENTRIES_PER_BLOCK = (MD_BLOCK_SIZE - sizeof(bitmap_header)) * 4;
@@ -230,23 +254,25 @@ namespace {
 		typedef transaction_manager::write_ref write_ref;
 
 		sm_disk(index_store::ptr indexes,
-			transaction_manager::ptr tm)
+			transaction_manager &tm)
 			: tm_(tm),
 			  bitmap_validator_(new bitmap_block_validator),
 			  indexes_(indexes),
 			  nr_blocks_(0),
 			  nr_allocated_(0),
+			  search_start_(0),
 			  ref_counts_(tm_, ref_count_traits::ref_counter()) {
 		}
 
 		sm_disk(index_store::ptr indexes,
-			transaction_manager::ptr tm,
+			transaction_manager &tm,
 			sm_root const &root)
 			: tm_(tm),
 			  bitmap_validator_(new bitmap_block_validator),
 			  indexes_(indexes),
 			  nr_blocks_(root.nr_blocks_),
 			  nr_allocated_(root.nr_allocated_),
+			  search_start_(0),
 			  ref_counts_(tm_, root.ref_count_root_, ref_count_traits::ref_counter()) {
 		}
 
@@ -266,50 +292,91 @@ namespace {
 			return count;
 		}
 
-		void set_count(block_address b, ref_t c) {
-			ref_t old = get_count(b);
+		template <typename Mut>
+		void modify_count(block_address b, Mut const &m) {
+			check_block(b);
 
-			if (c == old)
-				return;
+			index_entry ie = indexes_->find_ie(b / ENTRIES_PER_BLOCK);
+			bitmap bm(tm_, ie, bitmap_validator_);
+
+			ref_t old = bm.lookup(b % ENTRIES_PER_BLOCK);
+			if (old == 3)
+				old = lookup_ref_count(b);
+			ref_t c = m(old);
 
 			if (c > 2) {
-				if (old < 3)
-					insert_bitmap(b, 3);
+				if (old < 3) {
+					bm.insert(b % ENTRIES_PER_BLOCK, 3);
+					indexes_->save_ie(b / ENTRIES_PER_BLOCK, bm.get_ie());
+				}
 				insert_ref_count(b, c);
 			} else {
 				if (old > 2)
 					remove_ref_count(b);
-				insert_bitmap(b, c);
+				bm.insert(b % ENTRIES_PER_BLOCK, c);
+				indexes_->save_ie(b / ENTRIES_PER_BLOCK, bm.get_ie());
 			}
 
 			if (old == 0)
 				nr_allocated_++;
-			else if (c == 0)
+			else if (c == 0) {
+				if (b < search_start_)
+					search_start_ = b;
 				nr_allocated_--;
+			}
+		}
+
+		struct override {
+			override(ref_t new_value)
+				: new_value_(new_value) {
+				}
+
+			ref_t operator()(ref_t old) const {
+				return new_value_;
+			}
+
+			ref_t new_value_;
+		};
+
+		void set_count(block_address b, ref_t c) {
+			override m(c);
+			modify_count(b, m);
 		}
 
 		void commit() {
 			indexes_->commit_ies();
 		}
 
+		static ref_t inc_mutator(ref_t c) {
+			return c + 1;
+		}
+
+		static ref_t dec_mutator(ref_t c) {
+			return c - 1;
+		}
+
 		void inc(block_address b) {
-			// FIXME: 2 get_counts
-			ref_t old = get_count(b);
-			set_count(b, old + 1);
+			if (b == search_start_)
+				search_start_++;
+
+			modify_count(b, inc_mutator);
 		}
 
 		void dec(block_address b) {
-			ref_t old = get_count(b);
-			set_count(b, old - 1);
+			modify_count(b, dec_mutator);
 		}
 
-		// FIXME: keep track of the lowest free block so we
-		// can start searching from a suitable place.
 		maybe_block find_free(span_iterator &it) {
 			for (maybe_span ms = it.first(); ms; ms = it.next()) {
 				block_address begin = ms->first;
 				block_address end = ms->second;
 
+				if (end < search_start_)
+					continue;
+
+				if (begin < search_start_)
+					begin = search_start_;
+
 				block_address begin_index = begin / ENTRIES_PER_BLOCK;
 				block_address end_index = div_up<block_address>(end, ENTRIES_PER_BLOCK);
 
@@ -323,6 +390,8 @@ namespace {
 					boost::optional<unsigned> maybe_b = bm.find_free(bit_begin, bit_end);
 					if (maybe_b) {
 						block_address b = (index * ENTRIES_PER_BLOCK) + *maybe_b;
+						if (b)
+							search_start_ = b - 1;
 						return b;
 					}
 				}
@@ -343,7 +412,7 @@ namespace {
 
 			indexes_->resize(bitmap_count);
 			for (block_address i = old_bitmap_count; i < bitmap_count; i++) {
-				write_ref wr = tm_->new_block(bitmap_validator_);
+				write_ref wr = tm_.new_block(bitmap_validator_);
 
 				index_entry ie;
 				ie.blocknr_ = wr.get_location();
@@ -357,12 +426,14 @@ namespace {
 			nr_blocks_ = nr_blocks;
 		}
 
-		virtual void check(block_counter &counter) const {
-			ref_count_checker v(counter);
-			ref_counts_.visit_depth_first(v);
+		virtual void visit(space_map_detail::visitor &v) const {
+#if 0
+			ref_count_checker rcv(v);
+			ref_counts_.visit_depth_first(rcv);
 
 			block_address nr_entries = div_up<block_address>(get_nr_blocks(), ENTRIES_PER_BLOCK);
-			indexes_->check(counter, nr_entries);
+			indexes_->visit(v, nr_entries);
+#endif
 		}
 
 		struct look_aside_iterator : public iterator {
@@ -393,6 +464,13 @@ namespace {
 			}
 		}
 
+		virtual void count_metadata(block_counter &bc) const {
+			indexes_->count_metadata(bc);
+
+			noop_value_counter<uint32_t> vc;
+			count_btree_blocks(ref_counts_, bc, vc);
+		}
+
 		virtual size_t root_size() const {
 			return sizeof(sm_root_disk);
 		}
@@ -424,7 +502,7 @@ namespace {
 		}
 
 	protected:
-		transaction_manager::ptr get_tm() const {
+		transaction_manager &get_tm() const {
 			return tm_;
 		}
 
@@ -488,82 +566,84 @@ namespace {
 			ref_counts_.remove(key);
 		}
 
-		transaction_manager::ptr tm_;
-		block_manager<>::validator::ptr bitmap_validator_;
+		transaction_manager &tm_;
+		bcache::validator::ptr bitmap_validator_;
 		index_store::ptr indexes_;
 		block_address nr_blocks_;
 		block_address nr_allocated_;
+		block_address search_start_;
 
 		btree<1, ref_count_traits> ref_counts_;
 	};
 
-	class bitmap_tree_validator : public btree_checker<1, index_entry_traits> {
+	//--------------------------------
+
+	class ie_value_visitor {
 	public:
-		typedef btree_detail::node_location node_location;
-		typedef boost::shared_ptr<bitmap_tree_validator> ptr;
-
-		bitmap_tree_validator(block_counter &counter)
-			: btree_checker<1, index_entry_traits>(counter) {
+		ie_value_visitor(index_entry_visitor &v)
+			: v_(v) {
 		}
 
-		bool visit_leaf(node_location const &loc,
-				btree_detail::node_ref<index_entry_traits> const &n) {
-			bool r = btree_checker<1, index_entry_traits>::visit_leaf(loc, n);
-			if (!r)
-				return r;
-
-			for (unsigned i = 0; i < n.get_nr_entries(); i++) {
-				if (seen_indexes_.count(n.key_at(i)) > 0) {
-					ostringstream out;
-					out << "index entry " << i << " is present twice";
-					throw runtime_error(out.str());
-				}
-
-				seen_indexes_.insert(n.key_at(i));
-				btree_checker<1, index_entry_traits>::get_counter().inc(n.value_at(i).blocknr_);
-			}
-
-			return true;
-		}
-
-		void check_all_index_entries_present(block_address nr_entries) {
-			for (block_address i = 0; i < nr_entries; i++) {
-				if (seen_indexes_.count(i) == 0) {
-					ostringstream out;
-					out << "missing index entry " << i;
-					throw runtime_error(out.str());
-				}
-			}
-
-			set<block_address>::const_iterator it;
-			for (it = seen_indexes_.begin(); it != seen_indexes_.end(); ++it) {
-				if (*it >= nr_entries) {
-					ostringstream out;
-					out << "unexpected index entry " << *it;
-					throw runtime_error(out.str());
-				}
-			}
+		virtual void visit(btree_path const &path, sm_disk_detail::index_entry const &ie) {
+			// FIXME: finish
 		}
 
 	private:
-		set<block_address> seen_indexes_;
+		index_entry_visitor &v_;
+	};
+
+	class ie_damage_visitor {
+	public:
+		ie_damage_visitor(index_entry_visitor &v)
+			: v_(v) {
+		}
+
+		virtual void visit(btree_path const &path, btree_detail::damage const &d) {
+			// FIXME: finish
+		}
+
+	private:
+		index_entry_visitor &v_;
 	};
 
 	class btree_index_store : public index_store {
 	public:
 		typedef boost::shared_ptr<btree_index_store> ptr;
 
-		btree_index_store(transaction_manager::ptr tm)
+		btree_index_store(transaction_manager &tm)
 			: tm_(tm),
 			  bitmaps_(tm, index_entry_traits::ref_counter()) {
 		}
 
-		btree_index_store(transaction_manager::ptr tm,
+		btree_index_store(transaction_manager &tm,
 				  block_address root)
 			: tm_(tm),
 			  bitmaps_(tm, root, index_entry_traits::ref_counter()) {
 		}
 
+		//--------------------------------
+
+		struct index_entry_counter {
+			index_entry_counter(block_counter &bc)
+			: bc_(bc) {
+			}
+
+			void visit(btree_detail::node_location const &loc, index_entry const &ie) {
+				if (ie.blocknr_ != 0)
+					bc_.inc(ie.blocknr_);
+			}
+
+		private:
+			block_counter &bc_;
+		};
+
+		virtual void count_metadata(block_counter &bc) const {
+			index_entry_counter vc(bc);
+			count_btree_blocks(bitmaps_, bc, vc);
+		}
+
+		//--------------------------------
+
 		virtual void resize(block_address nr_entries) {
 			// No op
 		}
@@ -594,14 +674,14 @@ namespace {
 			return bitmaps_.get_root();
 		}
 
-		virtual void check(block_counter &counter, block_address nr_index_entries) const {
-			bitmap_tree_validator v(counter);
-			bitmaps_.visit_depth_first(v);
-			v.check_all_index_entries_present(nr_index_entries);
+		virtual void visit(index_entry_visitor &v, block_address nr_index_entries) const {
+			ie_value_visitor vv(v);
+			ie_damage_visitor dv(v);
+			btree_visit_values(bitmaps_, vv, dv);
 		}
 
 	private:
-		transaction_manager::ptr tm_;
+		transaction_manager &tm_;
 		btree<1, index_entry_traits> bitmaps_;
 	};
 
@@ -609,19 +689,30 @@ namespace {
 	public:
 		typedef boost::shared_ptr<metadata_index_store> ptr;
 
-		metadata_index_store(transaction_manager::ptr tm)
+		metadata_index_store(transaction_manager &tm)
 			: tm_(tm) {
-			block_manager<>::write_ref wr = tm_->new_block(index_validator());
+			block_manager<>::write_ref wr = tm_.new_block(index_validator());
 			bitmap_root_ = wr.get_location();
 		}
 
-		metadata_index_store(transaction_manager::ptr tm, block_address root, block_address nr_indexes)
+		metadata_index_store(transaction_manager &tm, block_address root, block_address nr_indexes)
 			: tm_(tm),
 			  bitmap_root_(root) {
 			resize(nr_indexes);
 			load_ies();
 		}
 
+		virtual void count_metadata(block_counter &bc) const {
+			bc.inc(bitmap_root_);
+
+			for (unsigned i = 0; i < entries_.size(); i++) {
+				block_address b = entries_[i].blocknr_;
+
+				if (b != 0)
+					bc.inc(b);
+			}
+		}
+
 		virtual void resize(block_address nr_indexes) {
 			entries_.resize(nr_indexes);
 		}
@@ -636,10 +727,10 @@ namespace {
 
 		virtual void commit_ies() {
 			std::pair<block_manager<>::write_ref, bool> p =
-				tm_->shadow(bitmap_root_, index_validator());
+				tm_.shadow(bitmap_root_, index_validator());
 
 			bitmap_root_ = p.first.get_location();
-			metadata_index *mdi = reinterpret_cast<metadata_index *>(&p.first.data());
+			metadata_index *mdi = reinterpret_cast<metadata_index *>(p.first.data());
 
 			for (unsigned i = 0; i < entries_.size(); i++)
 				index_entry_traits::pack(entries_[i], mdi->index[i]);
@@ -653,25 +744,31 @@ namespace {
 			return bitmap_root_;
 		}
 
-		virtual void check(block_counter &counter, block_address nr_index_entries) const {
+		virtual void visit(index_entry_visitor &vv, block_address nr_index_entries) const {
+			for (unsigned i = 0; i < entries_.size(); i++)
+				if (entries_[i].blocknr_ != 0)
+					vv.visit(entries_[i]);
+
+#if 0
 			counter.inc(bitmap_root_);
 			for (unsigned i = 0; i < entries_.size(); i++)
 				// FIXME: this looks like a hack
 				if (entries_[i].blocknr_ != 0) // superblock
 					counter.inc(entries_[i].blocknr_);
+#endif
 		}
 
 	private:
 		void load_ies() {
 			block_manager<>::read_ref rr =
-				tm_->read_lock(bitmap_root_, index_validator());
+				tm_.read_lock(bitmap_root_, index_validator());
 
-			metadata_index const *mdi = reinterpret_cast<metadata_index const *>(&rr.data());
+			metadata_index const *mdi = reinterpret_cast<metadata_index const *>(rr.data());
 			for (unsigned i = 0; i < entries_.size(); i++)
 				index_entry_traits::unpack(*(mdi->index + i), entries_[i]);
 		}
 
-		transaction_manager::ptr tm_;
+		transaction_manager &tm_;
 		block_address bitmap_root_;
 		std::vector<index_entry> entries_;
 	};
@@ -680,7 +777,7 @@ namespace {
 //----------------------------------------------------------------
 
 checked_space_map::ptr
-persistent_data::create_disk_sm(transaction_manager::ptr tm,
+persistent_data::create_disk_sm(transaction_manager &tm,
 				block_address nr_blocks)
 {
 	index_store::ptr store(new btree_index_store(tm));
@@ -691,7 +788,7 @@ persistent_data::create_disk_sm(transaction_manager::ptr tm,
 }
 
 checked_space_map::ptr
-persistent_data::open_disk_sm(transaction_manager::ptr tm, void *root)
+persistent_data::open_disk_sm(transaction_manager &tm, void const *root)
 {
 	sm_root_disk d;
 	sm_root v;
@@ -703,10 +800,15 @@ persistent_data::open_disk_sm(transaction_manager::ptr tm, void *root)
 }
 
 checked_space_map::ptr
-persistent_data::create_metadata_sm(transaction_manager::ptr tm, block_address nr_blocks)
+persistent_data::create_metadata_sm(transaction_manager &tm, block_address nr_blocks)
 {
 	index_store::ptr store(new metadata_index_store(tm));
 	checked_space_map::ptr sm(new sm_disk(store, tm));
+
+	if (nr_blocks > MAX_METADATA_BLOCKS) {
+		cerr << "truncating metadata device to " << MAX_METADATA_BLOCKS << " 4k blocks\n";
+		nr_blocks = MAX_METADATA_BLOCKS;
+	}
 	sm->extend(nr_blocks);
 	sm->commit();
 	return create_careful_alloc_sm(
@@ -714,7 +816,7 @@ persistent_data::create_metadata_sm(transaction_manager::ptr tm, block_address n
 }
 
 checked_space_map::ptr
-persistent_data::open_metadata_sm(transaction_manager::ptr tm, void *root)
+persistent_data::open_metadata_sm(transaction_manager &tm, void const *root)
 {
 	sm_root_disk d;
 	sm_root v;
@@ -728,4 +830,25 @@ persistent_data::open_metadata_sm(transaction_manager::ptr tm, void *root)
 			checked_space_map::ptr(new sm_disk(store, tm, v))));
 }
 
+bcache::validator::ptr
+persistent_data::bitmap_validator() {
+	return bcache::validator::ptr(new bitmap_block_validator());
+}
+
+bcache::validator::ptr
+persistent_data::index_validator() {
+	return bcache::validator::ptr(new index_block_validator());
+}
+
+block_address
+persistent_data::get_nr_blocks_in_data_sm(transaction_manager &tm, void *root)
+{
+	sm_root_disk d;
+	sm_root v;
+
+	::memcpy(&d, root, sizeof(d));
+	sm_root_traits::unpack(d, v);
+	return v.nr_blocks_;
+}
+
 //----------------------------------------------------------------
